import json
from ami_val.libs import utils_lib, resource_class, aws_lib

def test_stage0_check_aminame(test_instance):
    aminame = test_instance.info['name']
    if 'RHEL' in aminame:
        test_instance.log.info("RHEL is expected and found in AMI's name in push task")
        if 'SAP' in aminame:
            if 'Access2' in aminame:
                test_instance.fail('the -access images are not needed for this sap image set.(RHELDST-4739)')
    else:
        test_instance.fail('RHEL is expected in AMI name but get {} in push task'.format(aminame))
    test_instance.log.info("Details:{}".format(json.dumps(test_instance.info, indent=4)))

def test_stage0_check_ena_enabled(test_instance):
    if test_instance.info['ena_support'] != 'true':
        test_instance.log.info('ena_support is enabled in push task')
    else:
        test_instance.fail('ena_support is expected true but get {} in push task'.format(test_instance.info['ena_support']))
    test_instance.log.info("Details:{}".format(json.dumps(test_instance.info, indent=4)))

def test_stage0_launch_instance(test_instance):
    '''
    launch instances from AMIs in all supported regions
    '''
    aws_lib.aws_check_region(region=test_instance.info['region'], profile=test_instance.profile_name, resource_file=test_instance.resource_file, log=test_instance.log)
    vm = resource_class.EC2VM(test_instance)
    if vm.create():
        test_instance.vm = vm
        test_instance.ssh_client = vm.new_ssh_client()



import os, shutil
import random
import re
import time
import string
import logging
import decimal
import subprocess
import tipset
import json
import difflib
import urllib.request as request
import time
import importlib
import copy
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from filelock import FileLock
from . import aws_lib
from tipset.libs import minilog,rmt_ssh
from functools import wraps

def wait_for(ret=None, not_ret=None, ck_ret=False, ck_not_ret=False, timeout=60, interval=1, log=None):
    '''
    wait for a func return expected value within specified time
    '''
    if log == None:
        log = minilog.minilog()
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            while True:
                log.info("{} called, timeout {}".format(func.__name__, timeout))
                result = func(*args, **kwargs)
                if ck_ret and result == ret:
                    break
                if ck_not_ret and not_ret != result:
                    break
                end_time = time.time()
                time.sleep(interval)
                if end_time - start_time > timeout:
                    log.info('timeout, exit!')
                    break
            return result
        return wrapper
    return decorate

def write_sumlog(logfile, ami_id='', ami_name='', region='', casename='', result='', error=''):
    if not os.path.exists(logfile):
        with FileLock(logfile + '.lock'):
            with open(logfile, 'w+') as fh:
                x={}
                x['ami'] = ami_id
                x['result'] = [{"casename":casename, "result":result, "error":error}]
                x['region'] = region
                x['name'] = ami_name
                sums = [x]
                json.dump(sums, fh, indent=4)
        return True
    with FileLock(logfile + '.lock'):
        with open(logfile, 'r') as fh:
            sums = json.load(fh)
            is_found = False
            for ami in sums:
                if ami_id == ami['ami']:
                    is_found = True
                    break
            if is_found:
                for ami in sums:
                    if ami_id == ami['ami']:
                        ami['result'].append({"casename":casename, "result":result, "error":error})
                        break
            else:
                x={}
                x['ami'] = ami_id
                x['result'] = [{"casename":casename, "result":result, "error":error}]
                x['region'] = region
                x['name'] = ami_name
                sums.append(x)
        with open(logfile, 'w+') as fh:
            json.dump(sums, fh, indent=4)

def save_resource(logfile, ami_id=None, region=None, arch=None, instance_types_list=None, instance_type=None, instance_id=None, instance_hostname=None, subnet=None, sg=None, log=None):
    '''
    ami_id, region, arch is query condition and readonly when saving data to resource file.
    '''
    if log is None:
        log = minilog.minilog()
    if ami_id is None and region is None:
        log.info("Please specify ami_id or region to update resource file:{}".format(logfile))
        return True
    log.info("update resource file:{}".format(logfile))
    with FileLock(logfile + '.lock'):
        with open(logfile, 'r') as fh:
            sums = json.load(fh)
            is_found = False
            for ami in sums:
                #log.info("checking {} {} {}".format(ami['ami'], ami['region'], region))
                if ami_id is not None and ami_id == ami['ami']:
                    is_found = True
                    break
                if region is not None and region in ami['region']:
                    is_found = True
                    break
            if is_found:
                if ami_id is not None:
                    log.info("update ami info {}".format(ami_id))
                    for ami in sums:
                        if ami_id == ami['ami']:
                            if instance_type is not None:
                                ami['instance_type'] = instance_type
                            if instance_types_list is not None:
                                ami['instance_types_list'] = instance_types_list
                            if instance_id is not None:
                                ami['instance_id'] = instance_id
                            if instance_hostname is not None:
                                ami['instance_hostname'] = instance_hostname
                            if subnet is not None:
                                ami['subnet'] = subnet
                            if sg is not None:
                                ami['sg'] = sg
                if region is not None and arch is None:
                    log.info("update region info {}".format(region))
                    for ami in sums:
                        if region in ami['region']:
                            if instance_type is not None:
                                ami['instance_type'] = instance_type
                            if instance_types_list is not None:
                                ami['instance_types_list'] = instance_types_list
                            if instance_id is not None:
                                ami['instance_id'] = instance_id
                            if instance_hostname is not None:
                                ami['instance_hostname'] = instance_hostname
                            if subnet is not None:
                                ami['subnet'] = subnet
                            if sg is not None:
                                ami['sg'] = sg
                if region is not None and arch is not None:
                    log.info("update region info {} which arch {} matched".format(region, arch))
                    for ami in sums:
                        if region in ami['region'] and arch in ami["name"]:
                            if instance_type is not None:
                                ami['instance_type'] = instance_type
                            if instance_types_list is not None:
                                ami['instance_types_list'] = instance_types_list
                            if instance_id is not None:
                                ami['instance_id'] = instance_id
                            if instance_hostname is not None:
                                ami['instance_hostname'] = instance_hostname
                            if subnet is not None:
                                ami['subnet'] = subnet
                            if sg is not None:
                                ami['sg'] = sg
        with open(logfile, 'w+') as fh:
            json.dump(sums, fh, indent=4)
        log.info("updated resource file:{}".format(logfile))
    return True

def write_html(sumlog, region_missed=None, region_uploaded=None):
    """
    sumlog: result log
    region_missed: which region does not have this AMI
    region_uploaded: which region has this AMI
    """
    logfile = sumlog
    logdir = os.path.dirname(logfile)

    html_header = '''
<!DOCTYPE html>
<html lang="en">
<head>
  <title>AMI-VAL Test Report</title>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/4.5.2/css/bootstrap.min.css">
  <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.5.1/jquery.min.js"></script>
  <script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.16.0/umd/popper.min.js"></script>
  <script src="https://maxcdn.bootstrapcdn.com/bootstrap/4.5.2/js/bootstrap.min.js"></script>
  <script src="https://cdn.datatables.net/1.10.23/js/jquery.dataTables.min.js"></script>
  <script src="https://cdn.datatables.net/1.10.23/js/dataTables.bootstrap4.min.js"></script>
  <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/4.5.2/css/bootstrap.css">
  <link rel="stylesheet" href="https://cdn.datatables.net/1.10.23/css/dataTables.bootstrap4.min.css">
</head>
<body>

<div class="container-fluid">
  <h2>AMI-VAL Test Report</h2>
  <p>This report is for displaying RHEL AMIs validation result.</p>   
  <hr class="solid">       
    '''
    table_header = '''
     <table id="result" class="table table-hover table-bordered">
    <thead>
     <tr>
        <th>AMI</th>
        <th>Region</th>
        <th>Name</th>
        <th>CaseName</th>
        <th>Result</th>
        <th>Error</th>
        <th>Debug</th>
    </tr></thead>
    <tbody>
    '''
    html_tail = '''
    </tbody>
  </table>
</div>

</body>
</html>
<script>
$(document).ready(function() {
  $('#result').DataTable();
});
</script>
    '''
    with open(logdir + '/sum.html', 'a+') as fh:
        fh.write(html_header)
    summary = {'total':0, 'pass':0, 'fail':0, 'skip':0,'error':0}
    with open(logfile, 'r') as fh:
        sums = json.load(fh)
        for ami in sums:
            ami_id = ami['ami']
            region = ami['region']
            name = ami['name']
            for r in ami['result']:
                result = r['result']
                if 'PASS' in result:
                    summary['pass'] += 1
                elif 'FAIL' in result:
                    summary['fail'] += 1
                elif 'SKIP' in result:
                    summary['skip'] += 1
                elif 'ERROR' in result:
                    summary['error'] += 1
                else:
                    summary['fail'] += 1
                summary['total'] += 1
    pass_rate = (summary['pass'])/(summary['total'] - summary['skip'])*100
    summary = """
        <!-- Table -->
        <table class="table table-bordered">
            <tr>
            <td>Summary</td><td>Total:<b>{}</b>, PASS:<b class="text-success">{}</b>, FAIL:<b class="text-danger">{}</b>, SKIP:<b class="text-warning">{}</b>, ERROR:<b class="text-danger">{}</b> - success rate of <b>{}</b>% (excluding SKIP and CANCEL)</tr>
            <tr>
            <td>Regions Uploaded</td><td>{}</td></tr>
            <tr>
            <td>Regions Missed</td><td>{}</td></tr>
        </table>
    """.format(summary['total'],summary['pass'],summary['fail'],summary['skip'],summary['error'],format(pass_rate,'0.2f'),region_uploaded, region_missed)
    with open(logdir + '/sum.html', 'a+') as fh:
        fh.write(summary)
    with open(logdir + '/sum.html', 'a+') as fh:
        fh.write(table_header)
    with open(logfile, 'r') as fh:
        sums = json.load(fh)
        for ami in sums:
            ami_id = ami['ami']
            region = ami['region']
            name = ami['name']
            for r in ami['result']:
                casename = r['casename']
                result = r['result']
                error = r['error']
                debuglog = 'debug/' + ami_id + '_' + casename + '.log'
                item = "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td> \
                    <td><a href='{}'>debug</a></td></tr>".format(ami_id, region, name,
                    casename, result, error, debuglog)
                with open(logdir + '/sum.html', 'a+') as fh:
                    fh.write(item)
    with open(logdir + '/sum.html', 'a+') as fh:
        fh.write(html_tail)
            

def prepare_dir(args):
    logdir = args.logdir
    if 'http' in args.amis_file:
        print("Task url provided, try to download it")
        task_url = args.amis_file.replace('push','task')
        task_url = args.amis_file.replace('https','http')
        json_url = task_url + "/log/images.json?format=raw"
        print('Get data from %s' % json_url)
        s = request.urlopen(json_url)
        print('Got data from %s' % s.geturl())
        task_id = task_url.rstrip('/').split('/')[-1]
        logdir = "{}_{}".format(logdir, task_id)
        if os.path.exists(logdir):
            print("Remove exists {}".format(logdir))
            shutil.rmtree(logdir)
        if not os.path.exists(logdir):
            print("Create new {}".format(logdir))
            os.mkdir(logdir)
        amis_file = '{}/images.json'.format(logdir)
        if os.path.exists(amis_file):
            os.unlink(amis_file)
            print('Removed exists %s' % amis_file)
        with open(amis_file, 'b+w') as fh:
            fh.write(s.read())
        print('Data saved to %s' % amis_file)
    elif os.path.exists(args.amis_file):
        amis_file = '{}/images.json'.format(logdir)
        if os.path.exists(logdir):
            print("Remove exists {}".format(logdir))
            shutil.rmtree(logdir)
        if not os.path.exists(logdir):
            print("Create new {}".format(logdir))
            os.mkdir(logdir)
        shutil.copy(args.amis_file, amis_file )
    os.mkdir(logdir + "/debug")
    return logdir, amis_file

def tests_discover(args):
    this_dir = os.path.dirname(__file__)
    this_dir = os.path.dirname(this_dir)
    ts = ["ami_val.tests." + x for x in os.listdir(this_dir + '/tests') if x.startswith('test')]
    ts = [x[:-3] for x in ts]
    all_cases = []
    for stage in ts:
        # print("Run {} test".format(stage))
        #s = importlib.import_module(stage, __package__)
        s = importlib.import_module(stage)
        for case in dir(s):
            if case.startswith('test'):
                all_cases.append("{}.{}".format(s.__name__, case))
    #print("all cases{}".format(all_cases))
    if args.pattern is not None:
        print("case in stage0 is must required!")
        tmp_cases_list = []
        for case in all_cases:
            for p in args.pattern.split(','):
                if p in case or 'stage0' in case and case not in tmp_cases_list:
                    tmp_cases_list.append(case)
            if args.skip_pattern is not None and 'stage0' not in case:
                for p in args.skip_pattern.split(','):
                    if p in case and 'stage0' not in case:
                        if case in tmp_cases_list:
                            tmp_cases_list.remove(case)
    if args.pattern is None and args.skip_pattern is not None:
        print("case in stage0 is must required!")
        tmp_cases_list = copy.deepcopy(all_cases)
        for case in all_cases:
            if 'stage0' not in case:
                for p in args.skip_pattern.split(','):
                    if p in case and 'stage0' not in case:
                        if case in tmp_cases_list:
                            tmp_cases_list.remove(case)
    if args.pattern is not None or args.skip_pattern is not None:
        final_ts = tmp_cases_list
    else:
        final_ts = all_cases
    final_ts.sort()
    return final_ts


def msg_to_syslog(test_instance, msg=None):
    '''
    Save msg to journal log and dmesg.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        msg {string} -- msg want to save, default is casename
    Return:
        arm: return True
        other: return False
    '''
    if msg is None:
        msg = test_instance.id()
    cmd = "sudo echo os-tests:{} | systemd-cat -p info".format(msg)
    run_cmd(test_instance, cmd, expect_ret=0)
    cmd = "sudo bash -c 'echo \"{}\" > /dev/kmsg'".format(msg)
    run_cmd(test_instance, cmd, expect_ret=0)

def run_cmd(test_instance,
            cmd,
            expect_ret=None,
            expect_not_ret=None,
            expect_kw=None,
            expect_not_kw=None,
            expect_output=None,
            msg=None,
            cancel_kw=None,
            cancel_not_kw=None,
            cancel_ret=None,
            cancel_not_ret=None,
            timeout=60,
            ret_status=False,
            is_log_output=True,
            is_log_cmd=True,
            cursor=None,
            rmt_redirect_stdout=False,
            rmt_redirect_stderr=False,
            rmt_get_pty=False
            ):
    """run cmd with/without check return status/keywords and save log

    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        cmd {string} -- cmd to run
        expect_ret {int} -- expected return status
        expect_not_ret {int} -- unexpected return status
        expect_kw {string} -- string expected in output,seperate by ',' if
                              check multi words
        expect_not_kw {string} -- string not expected in output, seperate by
                                  ',' if check multi words
        expect_output {string} -- string exactly the same as output
        cancel_kw {string} -- cancel case if kw not found, seperate by ','
                              if check multi words
        cancel_not_kw {string} -- cancel case if kw found, seperate by ','
                              if check multi words
        cancel_ret {string} -- cancel case if ret code not match, seperate by ','
                              if check multi rets
        cancel_not_ret {string} -- cancel case if ret code found, seperate by ','
                              if check multi rets
        msg {string} -- addtional info to mark cmd run.
        ret_status {bool} -- return ret code instead of output
        is_log_output {bool} -- print cmd output or not
        is_log_cmd {bool} -- log cmd in plain text
        cursor {string} -- skip content before cursor(line)
        rmt_redirect_stdout {bool} -- ssh command not exit some times, redirect stdout to tmpfile if needed
        rmt_redirect_stderr {bool} -- ssh command not exit some times, redirect stderr to tmpfile if needed

    Keyword Arguments:
        check_ret {bool} -- [whether check return] (default: {False})
    """
    if msg is not None:
        test_instance.log.info(msg)
    if is_log_cmd:
        test_instance.log.info("CMD: {}".format(cmd))
    status = None
    output = None
    exception_hit = False

    try:
        if test_instance.ssh_client is not None:
            status, output = rmt_ssh.remote_excute(test_instance.ssh_client, cmd, timeout, redirect_stdout=rmt_redirect_stdout, redirect_stderr=rmt_redirect_stderr,rmt_get_pty=rmt_get_pty, log=test_instance.log)
        else:
            ret = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, timeout=timeout, encoding='utf-8')
            status = ret.returncode
            if ret.stdout is not None:
                output = ret.stdout
    except Exception as err:
        test_instance.log.error("Run cmd failed as {}".format(err))
        status = None
        exception_hit = True

    if exception_hit:
        test_instance.log.info("Try again")
        test_instance.log.info("Test via uname, if still fail, please make sure no hang or panic in sys")
        try:
            if test_instance.ssh_client is not None:
                status, output = rmt_ssh.remote_excute(test_instance.ssh_client, 'uname -a', timeout, log=test_instance.log)
                status, output = rmt_ssh.remote_excute(test_instance.ssh_client, cmd, timeout, redirect_stdout=rmt_redirect_stdout, redirect_stderr=rmt_redirect_stderr, rmt_get_pty=rmt_get_pty, log=test_instance.log)
            else:
                ret = subprocess.run('uname -a', shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, timeout=timeout, encoding='utf-8')
                status = ret.returncode
                if ret.stdout is not None:
                   output = ret.stdout
                test_instance.log.info("Return: {}".format(output.decode("utf-8")))
                ret = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, timeout=timeout, encoding='utf-8')
                status = ret.returncode
                if ret.stdout is not None:
                   output = ret.stdout
        except Exception as err:
            test_instance.log.error("Run cmd failed again {}".format(err))
    if cursor is not None and cursor in output:
        output = output[output.index(cursor):]
    if is_log_output:
        test_instance.log.info("CMD ret: {} out:{}".format(status, output))
    else:
        test_instance.log.info("CMD ret: {}".format(status))
    if expect_ret is not None:
         if status != expect_ret:
            test_instance.fail('ret is {}, expected ret is {}'.format(status, expect_ret))
    if expect_not_ret is not None:
        if status == expect_not_ret:
            test_instance.fail('ret is {}, expected not ret is {}'.format(status, expect_not_ret))
    if expect_kw is not None:
        for key_word in expect_kw.split(','):
            if output.count('\n') > 5:
                find_list = re.findall('\n.*{}.*\n'.format(key_word), output)
            else:
                find_list = re.findall('.*{}.*'.format(key_word), output)
            if len(find_list) > 0:
                test_instance.log.info('expcted "{}" found in "{}"'.format(key_word, ''.join(find_list)))
            else:
                if output.count('\n') > 5:
                    test_instance.fail('expcted "{}" not found in output(check debug log as too many lines)'.format(key_word))
                else:
                    test_instance.fail('expcted "{}" not found in "{}"'.format(key_word,output))
    if expect_not_kw is not None:
        for key_word in expect_not_kw.split(','):
            if output.count('\n') > 5:
                find_list = re.findall('\n.*{}.*\n'.format(key_word), output)
            else:
                find_list = re.findall('.*{}.*'.format(key_word), output)
            if len(find_list) == 0:
                test_instance.log.info('Unexpcted "{}" not found in output'.format(key_word))
            else:
                if output.count('\n') > 5:
                    test_instance.fail('Unexpcted "{}" found in {}'.format(key_word, ''.join(find_list)))
                else:
                    test_instance.fail('Unexpcted "{}" found in "{}"'.format(key_word,output))
    if expect_output is not None:
        if expect_output != output:
            test_instance.fail('exactly expected {}, but actual output is {}'.format(expect_output, output))

    if cancel_kw is not None:
        cancel_yes = True
        for key_word in cancel_kw.split(','):
            if key_word in output:
                cancel_yes = False
        if cancel_yes:
            test_instance.skipTest("'{}' not found, skip case. msg:{}".format(cancel_kw, msg))
    if cancel_not_kw is not None:
        for key_word in cancel_not_kw.split(','):
            if key_word in output:
                test_instance.skipTest("'{}' found, skip case. msg:{}".format(key_word, msg))
    if cancel_ret is not None:
        cancel_yes = True
        for ret in cancel_ret.split(','):
            if int(ret) == int(status):
                cancel_yes = False
        if cancel_yes:
            test_instance.skipTest("expected ret code '{}' not match ret '{}', cancel case. msg:{}".format(cancel_ret, status, msg))
    if cancel_not_ret is not None:
        for ret in cancel_not_ret.split(','):
            if int(ret) == int(status):
                test_instance.skipTest("skip ret code '{}' found, actual ret '{}' cancel case. msg:{}".format(ret, status, msg))
    if ret_status:
        return status
    return output

def compare_nums(test_instance, num1=None, num2=None, ratio=0, msg='Compare 2 nums'):
    '''
    Compare num1 and num2.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        num1 {int} -- num1
        num2 {int} -- num2
        ratio {int} -- allow ratio
    Return:
        num1 < num2: return True
        (num1 - num2)/num2*100 > ratio: return False
        (num1 - num2)/num2*100 < ratio: return True
    '''
    num1 = float(num1)
    num2 = float(num2)
    ratio = float(ratio)
    test_instance.log.info(msg)
    if num1 < num2:
        test_instance.log.info("{} less than {}".format(num1, num2))
        return True
    if (num1 - num2)/num2*100 > ratio:
        test_instance.fail("{} vs {} over {}%".format(num1, num2, ratio))
    else:
        test_instance.log.info("{} vs {} less {}%, pass".format(num1, num2, ratio))

def getboottime(test_instance):
    '''
    Get system boot time via "systemd-analyze"
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
    '''
    run_cmd(test_instance, "sudo which systemd-analyze", expect_ret=0)
    time_start = int(time.time())
    while True:
        output = run_cmd(test_instance, "sudo systemd-analyze")
        if 'Bootup is not yet finished' not in output:
            break
        time_end = int(time.time())
        run_cmd(test_instance, 'sudo systemctl list-jobs')
        if time_end - time_start > 60:
            test_instance.fail("Bootup is not yet finished after 60s")
        test_instance.log.info("Wait for bootup finish......")
        time.sleep(1)
    cmd = "sudo systemd-analyze blame > /tmp/blame.log"
    run_cmd(test_instance, cmd, expect_ret=0)
    run_cmd(test_instance, "cat /tmp/blame.log", expect_ret=0)
    output = run_cmd(test_instance, "sudo systemd-analyze", expect_ret=0)
    boot_time = re.findall("=.*s", output)[0]
    boot_time = boot_time.strip("=\n")
    boot_time_sec = re.findall('[0-9.]+s', boot_time)[0]
    boot_time_sec = boot_time_sec.strip('= s')
    if 'min' in boot_time:
        boot_time_min = re.findall('[0-9]+min', boot_time)[0]
        boot_time_min = boot_time_min.strip('min')
        boot_time_sec = int(boot_time_min) * 60 + decimal.Decimal(boot_time_sec).to_integral()
    test_instance.log.info(
        "Boot time is {}(s)".format(boot_time_sec))
    return boot_time_sec

def is_arch(test_instance, arch="", action=None):
    '''
    Check whether system is specific system.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        arch {string} -- arch want to check
        action {string} -- cancel case if it is not arch
    Return:
        arm: return True
        other: return False
    '''
    output = run_cmd(test_instance, "lscpu", expect_ret=0)
    if arch in output:
        test_instance.log.info("{}detected.".format(arch))
        return True
    else:
        if action == "cancel":
            test_instance.skipTest("Cancel it in non {} platform.".format(arch))
    test_instance.log.info("Not an {} instance.".format(arch))
    return False

def is_aws(test_instance, action=None):
    '''
    Check whether system is a aws system.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        action {string} -- cancel case if it is not a aws system
    Return:
        aws: return True
        other: return False
    '''
    output = run_cmd(test_instance, "cat /sys/devices/virtual/dmi/id/bios_*", expect_ret=0)
    if 'amazon' in output.lower():
        test_instance.log.info("AWS system.")
        return True
    else:
        if action == "cancel":
            test_instance.skipTest("Cancel it in non aws system.")
        test_instance.log.info("Not an aws system.")
    return False

def is_metal(test_instance, action=None):
    '''
    Check whether system is a baremetal system.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        action {string} -- cancel case if it is not a bare metal system
    Return:
        metal: return True
        other: return False
    '''
    output_lscpu = run_cmd(test_instance, "lscpu", expect_ret=0)
    if "x86_64" in output_lscpu and "Hypervisor" not in output_lscpu:
        test_instance.log.info("It is a bare metal instance.")
        return True
    elif "x86_64" in output_lscpu and "Hypervisor" in output_lscpu:
        test_instance.log.info("It is a virtual guest.")
        if action == "cancel":
            test_instance.skipTest("Cancel it in non metal system.")
        return False
    output_dmesg = run_cmd(test_instance, "dmesg", expect_ret=0, is_log_output=False)

    if 'HYP mode not available' in output_dmesg:
        test_instance.log.info("It is a virtual guest.")
        if action == "cancel":
            test_instance.skipTest("Cancel it in non metal system.")
        return False
    else:
        test_instance.log.info("It is a bare metal instance.")
        return True

def is_cmd_exist(test_instance, cmd=None, is_install=True, cancel_case=False):
    '''
    check cmd exists status, if no, try to install it.
    Arguments:
        test_instance {avocado Test instance} -- avocado test instance
        cmd {string} -- checked command
        is_install {bool} -- try to install it or not
    '''
    cmd_check = "which %s" % cmd
    ret = run_cmd(test_instance, cmd_check, ret_status=True)
    if ret == 0:
        return True
    else:
        test_instance.log.info("No %s found!" % cmd)
    if not is_install:
        if cancel_case:
            test_instance.skipTest("Cancel it as {} not found".format(cmd))
        return False
    arch = run_cmd(test_instance, 'uname -p').rstrip('\n')
    pkg_find = "sudo yum provides %s" % cmd
    output = run_cmd(test_instance, pkg_find, expect_ret=0)
    for i in [arch, 'noarch']:
        pkg_list = re.findall(".*%s" % i, output)
        if len(pkg_list) > 0:
            break
    if len(pkg_list) == 0:
        test_instance.skipTest("Unable to install {}".format(cmd))
        return False
    run_cmd(test_instance, "sudo yum install -y %s" % pkg_list[0], expect_ret=0, timeout=120)
    return True

def is_pkg_installed(test_instance, pkg_name=None, is_install=True, cancel_case=False):
    '''
    check cmd exists status, if no, try to install it.
    Arguments:
        test_instance {avocado Test instance} -- avocado test instance
        cmd {string} -- checked command
        is_install {bool} -- try to install it or not
    '''
    cmd_check = "rpm -qa|grep {}".format(pkg_name)
    ret = run_cmd(test_instance, cmd_check, ret_status=True)
    if ret == 0:
        return True
    else:
        test_instance.log.info("No %s found!" % pkg_name)
        return False

def pkg_install(test_instance, pkg_name=None, pkg_url=None):
        """
        Install pkg in target system from default repo or pkg_url.
        $pkg_url_$arch is defined in configuration file.
        I use pre compiled pkgs for saving time in run.
        eg.
        blktests_url_x86_64: https://github.com/liangxiao1/rpmbuild_specs/releases/download/blktests_20201009/blktests-master-20201009.aarch64.rpm
        or
        blktests_url_aarch64: https://github.com/liangxiao1/rpmbuild_specs/releases/download/blktests_20201009/blktests-master-20201009.aarch64.rpm
        Arguments:
            test_instance {avocado Test instance} -- avocado test instance
            pkg_name {string} -- pkg name
            pkg_url {string} -- pkg url if it is not in default repo
        """

        if not is_pkg_installed(test_instance, pkg_name=pkg_name):
            test_instance.log.info("Try install {} automatically!".format(pkg_name))
            if pkg_url is not None:
                test_instance.log.info("Install {} from {}".format(pkg_name, pkg_url))
                cmd = 'sudo yum -y install %s' % pkg_url
            else:
                test_instance.log.info("Install {} from default repo".format(pkg_name))
                cmd = 'sudo yum -y install %s' % pkg_name
            run_cmd(test_instance, cmd, timeout=1200)
        elif test_instance.params.get('pkg_reinstall'):
            test_instance.log.info("Try reinstall {} automatically!".format(pkg_name))
            if pkg_url is not None:
                test_instance.log.info("Reinstall {} from {}".format(pkg_name, pkg_url))
                cmd = 'sudo yum -y reinstall %s' % pkg_url
            else:
                test_instance.log.info("Reinstall {} from default repo".format(pkg_name))
                cmd = 'sudo yum -y reinstall %s' % pkg_name
            run_cmd(test_instance, cmd, timeout=1200)

        if not is_pkg_installed(test_instance, pkg_name=pkg_name) and pkg_url is not None:
            test_instance.log.info('Install without dependences!')
            cmd = 'sudo rpm -ivh %s --nodeps' % pkg_url
            run_cmd(test_instance, cmd, timeout=1200)
        if not is_pkg_installed(test_instance, pkg_name=pkg_name):
            test_instance.skipTest("Cannot install {} automatically!".format(pkg_name))

def get_memsize(test_instance, action=None):
    '''
    Check whether system is a aws system.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        action {string} -- cancel case if it is not a aws system
    Return:
        aws: return True
        other: return False
    '''
    output = run_cmd(test_instance, "cat /proc/meminfo |grep MemTotal", expect_ret=0)

    mem_kb = int(re.findall('\d+', output)[0])
    mem_gb = (mem_kb/1024/1024)
    test_instance.log.info("Total memory: {:0,.1f}GiB".format(mem_gb))
    return mem_gb

def get_cmd_cursor(test_instance, cmd='dmesg -T', rmt_redirect_stdout=False, rmt_get_pty=False):
    '''
    Get command cursor by last matched line.
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
    Return:
        cursor {string}
    '''
    output = run_cmd(test_instance, cmd, expect_ret=0, is_log_output=False, rmt_redirect_stdout=rmt_redirect_stdout, rmt_get_pty=rmt_get_pty)
    if len(output.split('\n')) < 5:
        return output.split('\n')[-1]
    for i in range(-1, -10, -1):
        cursor = output.split('\n')[i]
        if len(cursor) > 3:
            break
    test_instance.log.info("Get cursor: {}".format(cursor))
    return cursor

def check_log(test_instance, log_keyword, log_cmd="journalctl --since today", match_word_exact=False, cursor=None, skip_words=None, rmt_redirect_stdout=False, rmt_redirect_stderr=False, rmt_get_pty=False):
    '''
    check journal log
    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        log_keyword: which keywords to check, eg error, warn, fail, default is checking journal log happened in today.
        log_cmd: the command to get log
        match_word_exact: is macthing word exactly
        cursor: where to start to check journal log, only for journal log
        skip_words: skip words as you want, split by ","
    '''
    # Baseline data file
    baseline_file = os.path.dirname(os_tests.__file__) + "/data/baseline_log.json"
    # Result dir
    with open(baseline_file,'r') as fh:
        test_instance.log.info("Loading baseline data file from {}".format(baseline_file))
        baseline_dict = json.load(fh)
    run_cmd(test_instance, '\n')
    check_cmd = log_cmd

    if match_word_exact:
        check_cmd = check_cmd + '|grep -iw %s' % log_keyword
    ret = False
    if cursor is not None:
        out = run_cmd(test_instance,
                      check_cmd,
                      expect_ret=0,
                      msg='Get log......', cursor=cursor,
                      rmt_redirect_stderr=rmt_redirect_stderr,
                      rmt_redirect_stdout=rmt_redirect_stdout,
                      rmt_get_pty=rmt_get_pty)
    else:
        out = run_cmd(test_instance,
                      check_cmd,
                      expect_ret=0,
                      msg='Get log......',
                      rmt_redirect_stderr=rmt_redirect_stderr,
                      rmt_redirect_stdout=rmt_redirect_stdout,
                      rmt_get_pty=rmt_get_pty)

    for keyword in log_keyword.split(','):
        ret = find_word(test_instance, out, keyword, baseline_dict=baseline_dict, skip_words=skip_words)
        if not ret and baseline_dict is not None:
            test_instance.fail("New {} in {} log".format(keyword, check_cmd))
        elif not ret:
            test_instance.fail("Found {} in {} log!".format(keyword, check_cmd))
        else:
            test_instance.log.info("No unexpected {} in {} log!".format(keyword, check_cmd))

def clean_sentence(test_instance, line1, line2):
    """only keep neccessary words
    eg.
    line1: Sep 10 05:42:38 ip-172-31-1-196.us-west-2.compute.internal augenrules[783]: failure 1
    line2: augenrules[681]: failure 1

    return:
    line1: augenrules[681]: failure 1
    line2: augenrules[783]: failure 1

    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        line1 {string} -- string 1
        line2 {string} -- string 2
    Returns:
        line1
        line2
    """
    tmpline = ''
    line1_longer = True
    if len(line1) > len(line2):
        tmpline = line2
    else:
        tmpline = line1
        line1_longer = False
    for i in re.findall("\w+",tmpline):
        if len(i) >= 3:
            if i not in line1 or i not in line2:
                return line1, line2
            #test_instance.log.info("got start word {}".format(i))
            if line1_longer:
                line1 = line1[line1.index(i):]
            else:
                line2 = line2[line2.index(i):]
            #test_instance.log.info("return line1:{} line2:{}".format(line1, line2))
            return line1, line2
    return line1, line2

def find_word(test_instance, check_str, log_keyword, baseline_dict=None, skip_words=None):
    """find words in content

    Arguments:
        test_instance {Test instance} -- unittest.TestCase instance
        check_str {[string]} -- [string to look]
        baseline_dict {[dict]} -- [baseline dict to compare]
        match_word_exact: is macthing word exactly
        skip_words: skip words as you want, split by ","

    Returns:
        [Bool] -- [True|False]
    """
    tmp_list = re.findall('.*%s.*\n' % log_keyword, check_str, flags=re.I)
    if len(tmp_list) == 0:
        test_instance.log.info("No %s found!", log_keyword)
        return True
    else:
        test_instance.log.info("%s found!", log_keyword)
    if skip_words is not None:
        for skip_word in skip_words.split(','):
            tmp_list = [x for x in tmp_list if skip_word not in x]
    if len(tmp_list) == 0:
        test_instance.log.info("No {} found after skipped {}!".format(log_keyword, skip_words))
        return True
    # compare 2 string, if similary over fail_rate, consider it as same.
    fail_rate = 70
    no_fail = True
    for line1 in tmp_list:
        find_it = False
        if baseline_dict is not None:
            for basekey in baseline_dict:
                line1_tmp = line1
                line2_tmp = baseline_dict[basekey]["content"]
                line1_tmp, line2_tmp = clean_sentence(test_instance, line1_tmp, line2_tmp)
                seq = difflib.SequenceMatcher(
                    None, a=line1_tmp, b=line2_tmp)
                same_rate = seq.ratio() * 100
                if same_rate > fail_rate:
                    test_instance.log.info(
                        "Compare result rate: %d same, maybe it is not a \
new one", same_rate)
                    test_instance.log.info("Guest: %s Baseline: %s", line1,
                             baseline_dict[basekey]["content"])
                    test_instance.log.info("ID:%s Baseline analyze:%s Branch:%s Status:%s Link:%s Path:%s" %
                             (basekey,
                              baseline_dict[basekey]["analyze"],
                              baseline_dict[basekey]["branch"],
                              baseline_dict[basekey]["status"],
                              baseline_dict[basekey]["link"],
                              baseline_dict[basekey]["path"]))
                    if baseline_dict[basekey]["trigger"] in check_str and len(baseline_dict[basekey]["trigger"]) > 2:
                        test_instance.log.info("Maybe it is expected because found '{}' too".format(baseline_dict[basekey]["trigger"]))
                        find_it = True
                    if baseline_dict[basekey]["status"] == 'active':
                        find_it = True
                    else:
                        test_instance.log.info("Find a similar issue which should be already fixed, please check manually.")
                        find_it = False
                        no_fail = False
                    break
        if not find_it:
            test_instance.log.info("This is a new exception!")
            test_instance.log.info("{}".format(line1))
            no_fail = False

    return no_fail



from . import utils_lib

import boto3
import time
from botocore.exceptions import ClientError
from botocore.config import Config
import botocore.exceptions as boto_err
from tipset.libs import minilog, rmt_ssh
import json
import ami_val

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import os
from . import utils_lib
from filelock import FileLock

class FailException(Exception):
    pass
class SkipException(Exception):
    pass

class BaseTest():
    def __init__(self):
        self.info = None
        self.params = None
        self.vm = None
        self.logdir = None
        self.debuglog = None
        self.casename = None
        self.sumlog = None
        self.ssh_username = None
        self.ssh_keyfile = None
        self.keypairname = None
        self.instance_type = None
        self.subnet = None
        self.security_group = None
        self.tag_name = None
        self.ssh_conn = None
        self.instance_ip = None
        self.log = minilog.minilog()
        self.result = 'PASS'
        self.error = ''
        self.profile_name = 'aws'
        self.ssh_client = None
        self.resource_file = None
        self.resource_info = None

    def load_resource(self):
        self.resource_file = self.logdir + '/resource.json'
        with FileLock(self.resource_file + '.lock'):
            with open(self.resource_file,'r') as fh:
                tmp_resource = json.load(fh)
                for r in tmp_resource:
                    if r['ami'] == self.info['ami']:
                        self.resource_info = r
                        break

    def sum_log(self, result='', error=''):
        utils_lib.write_sumlog(self.sumlog,ami_id=self.info['ami'], ami_name=self.info['name'], region=self.info['region'], casename=self.casename, result=result, error=error)

    def fail(self, msg, is_raise=True):
        self.log.error(msg)
        self.error = msg
        self.result = 'FAIL'
        if is_raise:
            raise FailException(self.error)

    def skipTest(self, msg, is_raise=True):
        self.log.info(msg)
        self.error = msg
        self.result = 'SKIP'
        if is_raise:
            raise SkipException(self.error)

class EC2VM():
    __ec2_instance = None

    def __init__(self, testinstance):
        self.log = testinstance.log
        config = Config(retries=dict(max_attempts=2, ))
        self.profile_name = testinstance.profile_name
        if self.profile_name is None:
            self.profile_name = 'default'
        self.log.info('Load profile_name: {}'.format(self.profile_name))
        self.session = boto3.session.Session(profile_name=self.profile_name, region_name=testinstance.info.get('region'))
        self.__resource = self.session.resource('ec2', config=config)
        self.__client = self.session.client('ec2', config=config, region_name=testinstance.info.get('region'))
        #super(EC2VM, self).__init__(testinstance)
        self.instance_id = None
        self.ipv4 = None
        self.ami_id = testinstance.info['ami']
        self.resource_file = testinstance.logdir + '/resource.json'
        with FileLock(self.resource_file + '.lock'):
            with open(self.resource_file,'r') as fh:
                tmp_resource = json.load(fh)
                for r in tmp_resource:
                    if r['ami'] == self.ami_id:
                        self.resource = r
                        break  
        self.instance_type = None
        self.log.info("AMI picked {}, instance type {}".format(self.ami_id, self.instance_type))
        self.subnet = self.resource['subnet']
        self.sg = self.resource['sg']
        self.log.info("subnet {}, sg {}".format(self.subnet, self.sg))

        self.region = testinstance.info.get('region')
        # Config file
        cfg_file = os.path.dirname(ami_val.__file__) + "/cfg/ami-val.yaml"
        # Result dir
        with open(cfg_file,'r') as fh:
            keys_data = load(fh, Loader=Loader)
        self.tagname = keys_data['tag_name']

        self.ssh_key_name = keys_data['pair_name']
        self.ssh_key_path = keys_data['ssh_keyfile']
        self.ssh_username = keys_data['ssh_user']
        self.ssh_conn = None
        self.__volume_id = None
        self.is_created = False
        self.another_ip = None
        self.additionalinfo = None

    def reuse_init(self, instance_id):
        '''
        To reuse an exist instance than create a new one
        @params: instance_id id of existing instance
        '''
        self.log.info("reuse instance: {}".format(instance_id))
        if instance_id is None:
            return False
        try:
            self.__ec2_instance = self.__resource.Instance(instance_id)
            self.__ec2_instance.reload()
            self.log.info("instance state: {}".format(self.__ec2_instance.state))
            for x in range(10):
                if self.is_stopping():
                    self.log.info("Wait for 60 seconds, max 10 mins")
                    time.sleep(60)
                else:
                    break
            if self.is_stopping():
                self.log.info("Instance is still stopping, cannot use it ")
                return False
            if self.is_shutting_down():
                self.log.info("Instance is shutting-down, cannot reuse it")
                return False
            if self.is_deleted():
                self.log.info("Instance is terminated, cannot reuse it")
                return False
            self.log.info("Reuse {} without instance type check!".format(instance_id))
            self.is_created = False
            return True
        except Exception as err:
            self.log.info("reuse init failed: {}".format(err))
            return False

    def create(self, wait=True):
        self.is_created = False
        if self.additionalinfo == None or self.additionalinfo == '':
            if self.instance_type is None and len(self.resource['instance_types_list']) == 0:
                self.log.info("No instance type specified and instance_types_list is empty")
                return self.is_created
            for instance_type in self.resource['instance_types_list']:
                if self.instance_type is not None:
                    instance_type = self.instance_type
                self.log.info("try to create {}".format(instance_type))
                try:
                    self.__ec2_instance = self.__resource.create_instances(
                        ImageId=self.ami_id,
                        InstanceType=instance_type,
                        KeyName=self.ssh_key_name,
                        #SecurityGroupIds=[
                        #    self.security_group_ids,
                        #],
                        #SubnetId=self.subnet_id,
                        MaxCount=1,
                        MinCount=1,
                        #Placement={
                        #    'AvailabilityZone': self.zone,
                        #},
                        NetworkInterfaces=[
                            {
                                'AssociatePublicIpAddress': True,
                                'DeviceIndex': 0,
                                'SubnetId': self.subnet,
                                'Groups': [
                                     self.sg,
                                 ],
                            },
                        ],
                        TagSpecifications=[
                            {
                                'ResourceType': 'instance',
                                'Tags': [
                                    {
                                        'Key': 'Name',
                                        'Value': self.tagname
                                    },
                                ]
                            },
                        ]
                        )
                        #UserData='#!/bin/bash\nmkdir /home/%s/instance_create_%s' %
                        #(self.ssh_user, self.instance_type))[0]
                    self.is_created = True 
                    break
                except ClientError as err:
                    self.log.info("Failed to create instance!")
                    self.is_createable = False
                    self.log.info(err)
                except Exception as err:
                    self.log.info(err)

        if self.additionalinfo != None and self.additionalinfo != '':
            for additionalinfo in self.additionalinfo.split(';'):
                try:
                    self.log.info("Create instance using addtionalinfo: {}".format(additionalinfo))
                    self.__ec2_instance = self.__resource.create_instances(
                        ImageId=self.ami_id,
                        InstanceType=self.instance_type,
                        KeyName=self.ssh_key_name,
                        #SecurityGroupIds=[
                        #    self.security_group_ids,
                        #],
                        #SubnetId=self.subnet_id,
                        MaxCount=1,
                        MinCount=1,
                        AdditionalInfo=additionalinfo,
                        NetworkInterfaces=[
                        {
                            'AssociatePublicIpAddress': True,
                            'DeviceIndex': 0,
                            'SubnetId': self.subnet,
                            'Groups': [
                                 self.sg,
                             ],
                        },
                    ],
                        UserData='#!/bin/bash\nmkdir /home/%s/instance_create_%s' %
                        (self.ssh_user, self.instance_type))[0]
                    self.is_created = True
                except ClientError as err:
                    self.log.info("Failed to create instance, try another addtionalinfo {}".format(err))
                except Exception as err:
                    self.log.info("Failed to create instance, try another addtionalinfo {}".format(err))
                if self.is_created:
                    break
            if not self.is_created:
                raise err

        #self.create_tags()
        self.log.info(self.__ec2_instance)
        if self.is_created:
            self.__ec2_instance = self.__ec2_instance[0]
        else:
            return self.is_created
        if wait:
            try:
                self.__ec2_instance.wait_until_running()
                self.__ec2_instance.reload()
            except Exception as err:
                self.log.info("Failed to wait instance running!{}".format(err))

        self.instance_id = self.__ec2_instance.id
        # self.ipv4 = self.__ec2_instance.public_ip_address
        self.floating_ip
        self.boot_volume_id
        utils_lib.save_resource(self.resource_file, ami_id=self.ami_id, instance_id=self.__ec2_instance.id, log=self.log)
        return True

    def create_tags(self):
        try:
            self.__client.create_tags(Resources=[self.__ec2_instance.id],
                                      Tags=[{
                                          'Key': 'Name',
                                          'Value': self.tagname
                                      }])
            self.log.info("Added tag: {} to instance: {}".format(self.tagname, self.__ec2_instance.id))
        except ClientError as err:
            self.log.info("Failed to add tag to {}".format(self.__ec2_instance.id))

    def show(self):
        pass

    @property
    def res_id(self):
        '''
        return resource id for local resource management
        '''
        return self.__ec2_instance.id

    @property
    def res_type(self):
        '''
        return resource id for local resource management
        '''
        return self.__ec2_instance.instance_type

    @property
    def res_name(self):
        '''
        return resource name for local resource management
        '''
        return 'instance'

    def start(self, wait=True):
        try:
            self.__ec2_instance.start()
        except Exception as err:
            self.log.info(err)
            return False

        if wait:
            self.__ec2_instance.wait_until_running()
            if self.__ec2_instance.state['Name'] == 'running':
                self.log.info("Instance is in running state!")
            else:
                self.log.info(
                    "Instance is not in running state! It is in {} state!".format(self.__ec2_instance.state['Name']))
                return False
            self.__ec2_instance.reload()
            self.floating_ip
        return True

    def stop(self, wait=True, loops=4):
        try:
            self.log.info("Stopping instance {} ".format(self.instance_id))
            self.__ec2_instance.stop()
        except Exception as err:
            self.log.info("{}".format(err))
            return False

        if wait:
            for i in range(0, loops):
                self.log.info("Wait loop {}, max loop {}".format(i, loops))
                try:
                    self.__ec2_instance.wait_until_stopped()
                    return self.is_stopped()
                except boto_err.WaiterError as err:
                    self.log.info("{}".format(err))
            return self.is_stopped()
        return True

    def reboot(self, wait=False):
        '''
        Reboot from outside
        '''
        self.log.info("Rebooting instance: {}".format(self.instance_id))
        try:
            self.__ec2_instance.reboot()
            if 'metal' in self.instance_type:
                time.sleep(120)
            else:
                time.sleep(5)
            return True
        except Exception as err:
            self.log.info(err)
            return False

    def delete(self, wait=True, loops=4):
        try:
            self.log.info("Deleting instance: {}".format(self.__ec2_instance.id))
            self.__ec2_instance.terminate()
        except Exception as err:
            self.log.info(err)
            return False
        if wait:
            for i in range(0, loops):
                self.log.info("Wait loop {}, max loop {}".format(i, loops))
                try:
                    self.__ec2_instance.wait_until_terminated()
                    return self.is_deleted()
                except boto_err.WaiterError as err:
                    self.log.info(err)
            return self.is_deleted()
        return True

    def send_nmi(self):
        try:
            self.log.info("Send diagnostic interrupt to {}".format(self.__ec2_instance.id))
            self.__client.send_diagnostic_interrupt(
                InstanceId=self.__ec2_instance.id, DryRun=False)
            return True
        except ClientError as err:
            self.log.info("Failed to send_diagnostic_interrupt to {}".format(self.__ec2_instance.id))
            return False

    def exists(self):
        if self.__ec2_instance is None:
            self.log.info("Instance does not exist!")
            return False
        if self.is_deleted():
            self.log.info("Instance is in terminate state!")
            return False
        else:
            self.log.info("Instance exists: {}".format(self.instance_id))
            return True

    def is_started(self):
        try:
            self.__ec2_instance.reload()
            if self.__ec2_instance.state['Name'] == 'running':
                self.log.info("Instance is in running state!")
                return True
            else:
                self.log.info(
                    "Instance is not in running state! It is in {} state".format(self.__ec2_instance.state['Name']))
                return False
        except Exception as err:
            self.log.info(err)
            return False

    def is_stopped(self):
        self.__ec2_instance.reload()
        if self.__ec2_instance.state['Name'] == 'stopped':
            self.log.info("Instance is in stopped state!")
            return True
        else:
            self.log.info("Instance is not in stopped state! It is in {} state!".format(self.__ec2_instance.state['Name']))
            return False

    def is_stopping(self):
        self.__ec2_instance.reload()
        if self.__ec2_instance.state['Name'] == 'stopping':
            self.log.info("Instance is in stopping state!")
            return True
        else:
            self.log.info("Instance is not in stopping state! It is in {} state!".format(self.__ec2_instance.state['Name']))
            return False

    def is_shutting_down(self):
        self.__ec2_instance.reload()
        if self.__ec2_instance.state['Name'] == 'shutting-down':
            self.log.info("Instance is in shutting-down state!")
            return True
        else:
            self.log.info(
                "Instance is not in shutting-down state! It is in {} state!".format(self.__ec2_instance.state['Name']))
            return False

    def is_deleted(self):
        try:
            self.__ec2_instance.reload()
            if self.__ec2_instance.state['Name'] == 'terminated':
                self.log.info("Instance is in terminated state!")
                return True
            else:
                self.log.info(
                    "Instance is not in terminated state! It is in {} state!".format(self.__ec2_instance.state['Name']))
                return False
        except Exception as err:
            self.log.info("Failed to get instance status, it may not exist! {}".format(err))     
            return True

    @property
    #@utils_lib.wait_for(not_ret=None, ck_not_ret=True, timeout=120)
    def floating_ip(self, interval=2, timeout=90):
        start_time = time.time()
        while True:
            self.log.info("reload and get public ip")
            self.__ec2_instance.reload()
            self.ipv4 = self.__ec2_instance.public_dns_name
            if self.ipv4 is not None:
                break
            end_time = time.time()
            time.sleep(interval)
            if end_time - start_time > timeout:
                log.info('timeout, break!')
                break
        self.log.info("Public ip is: {}".format(self.ipv4))
        return self.ipv4

    @property
    def priviate_ip(self):
        self.__ec2_instance.reload()
        self.log.info("Private ip is: {}".format(self.__ec2_instance.private_ip_address))
        return self.__ec2_instance.private_ip_address

    @property
    def boot_volume_id(self):
        for i in self.__ec2_instance.volumes.all():
            if 'sda' in i.attachments[0].get('Device') \
                    or 'xvda' in i.attachments[0].get('Device') \
                    or 'nvme0' in i.attachments[0].get('Device'):
                self.__volume_id = i.id
                self.log.info("Boot volume id: {}".format(self.__volume_id))
                return self.__volume_id
        return None

    @property
    def primary_nic_id(self):
        for nic in self.__ec2_instance.network_interfaces_attribute:
            if nic['Attachment']['DeviceIndex'] == 0:
                return nic['NetworkInterfaceId']
        self.log.info("primary nic id not found")
        return None

    def assign_new_ip(self):
        nic = self.__resource.NetworkInterface(self.primary_nic_id)
        ret = nic.assign_private_ip_addresses(
                AllowReassignment=True,
                SecondaryPrivateIpAddressCount=1
            )
        self.another_ip = ret['AssignedPrivateIpAddresses'][0]['PrivateIpAddress']
        self.log.info("second nic ip{}".format(self.another_ip))
        return self.another_ip
    
    def remove_added_ip(self):
        nic = self.__resource.NetworkInterface(self.primary_nic_id)
        if self.another_ip is None:
            self.log.info("second nic ip is {}".format(self.another_ip))
            return False
        try:
            ret = nic.unassign_private_ip_addresses(
                    PrivateIpAddresses=[
                        self.another_ip,
                    ]
                )
            self.log.info("removed second nic ip{}".format(self.another_ip))
            return True
        except Exception as err:
            self.log.info(err)
        return False

    def get_volumes_id(self):
        volumes_list = []
        self.log.info("Try to get all attached volumes!")
        self.__ec2_instance.reload()
        for i in self.__ec2_instance.volumes.all():
            volumes_list.append(i.id)
        self.log.info(volumes_list)
        return volumes_list

    def get_console_log(self):
        try:
            output = self.__ec2_instance.console_output(Latest=True).get('Output')
            return True, output
        except Exception as err:
            self.log.info("Failed to get console log, try without latest! {}".format(err))
        try:
            output = self.__ec2_instance.console_output().get('Output')
            return True, output
        except Exception as err:
            self.log.info("Failed to get console log! {}".format(err))
            return False, err

    def modify_instance_type(self, new_type):
        try:
            self.__ec2_instance.modify_attribute(
                InstanceType={'Value': new_type})
            self.instance_type = new_type
            return True
        except Exception as err:
            self.log.info("Failed to change instance type to {}, ret:{}".format(new_type, err)) 
            return False
    
    def new_ssh_client(self):
        return rmt_ssh.build_connection(rmt_node=self.floating_ip, rmt_user=self.ssh_username, rmt_password=None, rmt_keyfile=self.ssh_key_path, timeout=180, log=self.log)



import os
import sys
import time
import json
from ami_val.libs.utils_lib import run_cmd
import ami_val

def test_stage2_check_ha_specific(test_instance):
    if 'HA' not in test_instance.info['name']:
        test_instance.skipTest('only run in HA AMIs')
    else:
        script_dir = os.path.realpath(ami_val.__file__)
        script_dir = os.path.dirname(script_dir)
        script_file = script_dir + '/scripts/rhel-ha-aws-check.sh'
        rmt_file = '/tmp/rhel-ha-aws-check.sh'
        ftp_client = test_instance.ssh_client.open_sftp()
        ftp_client.put(script_file, rmt_file)
        run_cmd(test_instance, 'sudo chmod 555 {}'.format(rmt_file), msg='make it executeable')
        run_cmd(test_instance, 'sudo {} 2>&1'.format(rmt_file), expect_ret=0, msg='run ha test', timeout=1800)

def test_stage2_check_libc6_xen_conf(test_instance):
    """
    check for /etc/ld.so.conf.d/libc6-xen.conf absence on RHEL
    """
    run_cmd(test_instance, 'sudo test -f /etc/ld.so.conf.d/libc6-xen.conf', expect_ret=1, msg='check for /etc/ld.so.conf.d/libc6-xen.conf absence on RHEL')


import sys
import os
import logging
import ami_val
import json
import copy
try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    print("Please install boto3")
    sys.exit(1)

from tipset.libs import minilog
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import concurrent.futures
from . import utils_lib

def aws_init_key(region=None, profile=None, log=None):
    if log is None:
        log = minilog.minilog()
    default_regions = ['us-west-2', 'cn-northwest-1', 'us-gov-west-1']
    if region is None:
        default_regions = ['us-west-2', 'cn-northwest-1', 'us-gov-west-1']
    else:
        default_regions = [region]
    if profile is None:
        profile = 'default'
    log.info("Use profile:{} in region {}".format(profile,default_regions))
    for region in default_regions:
        try:
            session = boto3.session.Session(profile_name=profile, region_name=region)
            ec2_resource = session.resource('ec2', region_name=region)
            ec2_client = session.client('ec2', region_name=region)
            region_list = ec2_client.describe_regions()['Regions']
            log.info("Init key in region {} successfully".format(region))
            break
        except Exception as error:
            log.info("Try to init in region:{} result:{}".format(region,str(error)))
    return ec2_resource, ec2_client

def aws_vpc_sg_find(vpcid, region, profile_name, log=None):
    '''
    if the vpc's default security group allow ssh connection
    return None or sg id
    '''
    if log is None:
        log = minilog.minilog()
    ec2, _ = aws_init_key(region,profile=profile_name, log=log)
    try:
        vpc = ec2.Vpc(vpcid)
        log.info("vpc init {}".format(vpcid))
    except Exception as error:
        log.info("vpc init error {}, {}", vpcid, str(error))
        return None
    try:
        sgs = vpc.security_groups.all()
    except Exception as error:
        log.info("default sg get error: {}".format(str(error)))
        return None
    for sg in sgs:
        try:
            sg = ec2.SecurityGroup(sg.id)
            ips = sg.ip_permissions
            for ip in ips:
                ip_ranges = ip['IpRanges']
                log.debug(ip_ranges)
                for ip_range in ip_ranges:
                    if '0.0.0.0/0' in ip_range['CidrIp']:
                        log.info("found security group:{} vpc check pass!".format(sg.id))
                        security_group_id = sg.id
                        return security_group_id
            log.info("ssh enabled security group not found in {}".format(vpcid))
        except Exception as error:
            log.info("sg init error {}, error:{}".format(sg.id, error))
    return None

def aws_vpc_create(region, profile_name, tag, log=None):
    '''
    create a new vpc for test running
    '''
    if log is None:
        log = minilog.minilog()
    log.info("create a new vpc for test running")
    ec2, client = aws_init_key(region,profile=profile_name, log=log)
    default_tag = 'virtqe'
    if tag is None:
        tag = default_tag
    try:
        vpc_new = client.create_vpc(
            CidrBlock='192.111.0.0/16',
            AmazonProvidedIpv6CidrBlock=True,
            DryRun=False,
            InstanceTenancy='default'
        )
    except Exception as err:
        log.info("Failed to create vpc {}".format(str(err)))
    vpcid = vpc_new['Vpc']['VpcId']
    log.info("New vpc created {}".format(vpcid))
    try:
        vpc = ec2.Vpc(vpcid)
        log.info("vpc init {}".format(vpcid))
        tag = vpc.create_tags(
            DryRun=False,
            Tags=[
                {
                    'Key': 'Name',
                    'Value': tag
                },
            ]
        )
        log.info("added tag to vpc: {}".format(tag))
        vpc.modify_attribute(
            EnableDnsHostnames={
                'Value': True
            }
        )
        log.info("Enabled dns support")
    except Exception as error:
        log.info(str(error))
        return False
    igw = aws_igw_create(client, vpcid, tag)
    if igw == None:
        return vpc
    rt = aws_rt_update(client, vpc, igw, tag)
    if rt == None:
        return vpc
    subnet = aws_subnet_create(client, vpc, tag)
    if subnet == None:
        return vpc
    sg = aws_sg_update(client, vpc, igw, tag)
    if sg == None:
        return vpc

def aws_subnet_find(region, profile, create_new=True, log=None):
    '''
    if subnet enable public ipv4 on launch.
    return None or subnet id
    '''
    if log is None:
        log = minilog.minilog()
    _, client = aws_init_key(region,profile=profile, log=log)
    subnet_id = None
    sg_id = None
    subnets = client.describe_subnets()['Subnets']
    for subnet in subnets:
        if subnet['MapPublicIpOnLaunch']:
            vpc_id = subnet['VpcId']
            sg_id = aws_vpc_sg_find(vpc_id, region, profile, log=log)
            if sg_id:
                subnet_id = subnet['SubnetId']
                break
    if subnet_id is None and create_new:
        log.info("No ipv4 pub enabed subnets found in region {}".format(region))
        vpc = aws_vpc_create(region, profile, 'virtqe')
        subnets = client.describe_subnets()['Subnets']
        for subnet in subnets:
            if subnet['MapPublicIpOnLaunch']:
                vpc_id = subnet['VpcId']
                sg_id = aws_vpc_sg_find(vpc_id, region, profile, log=log)
                if sg_id:
                    subnet_id = subnet['SubnetId']
                    break
    if subnet_id is None:
        log.info("No proper subnet(ssh enabled) found in {}, please check manually!".format(region))
    log.info("Found existing subnet: {} in region {}".format(subnet_id, region))
    return subnet_id, sg_id, region

def aws_igw_create(region, profile, vpcid, tag, log=None):
    '''
    create a new igw and attach to vpc
    '''
    if log is None:
        log = minilog.minilog()
    ec2, client = aws_init_key(region,profile=profile, log=log)
    default_tag = 'virtqe'
    if tag is None:
        tag = default_tag
    try:
        igw_new = client.create_internet_gateway(
            DryRun=False
        )
        igwid = igw_new['InternetGateway']['InternetGatewayId']
        log.info("New igw created {}".format(igwid))
        igw = ec2.InternetGateway(igwid)
        igw.create_tags(
            DryRun=False,
            Tags=[
                {
                    'Key': 'Name',
                    'Value': tag
                },
            ]
        )
        igw.attach_to_vpc(
            DryRun=False,
            VpcId=vpcid
        )
        return igw
    except Exception as err:
        if 'Resource.AlreadyAssociated' in str(err):
            return igw
        log.info(str(err))
        return None

def aws_rt_update(region, profile, vpcid, igwid, tag, log=None):
    '''
    update default route table
    '''
    if log is None:
        log = minilog.minilog()
    ec2, client = aws_init_key(region,profile=profile, log=log)
    default_tag = 'virtqe'
    if tag is None:
        tag = default_tag
    try:
        vpc = ec2.Vpc(vpcid)
        igw = ec2.InternetGateway(igwid)
        rts = vpc.route_tables.all()
        for i in rts:
            for x in i.associations_attribute:
                if x['Main']:
                    log.info("found route table, {}".format(i.id))
                    rt = i
        #rt = vpc.create_route_table(
        #    DryRun=False,
        #
        #)
        #log.info("New route table created %s", rt.id)
        log.info("Update route table {}".format(rt.id))
        rt.create_tags(
            DryRun=False,
            Tags=[
                {
                    'Key': 'Name',
                    'Value': tag
                },
            ]
        )
        log.info("tag added")
        route = rt.create_route(
            DestinationCidrBlock='0.0.0.0/0',
            DryRun=False,
            GatewayId=igw.id,
        )
        return rt
    except Exception as err:
        log.info(str(err))
        return None

def aws_sg_update(region, profile, vpcid, tag, log=None):
    '''
    update default security group
    '''
    if log is None:
        log = minilog.minilog()
    ec2, client = aws_init_key(region,profile=profile, log=log)
    default_tag = 'virtqe'
    if tag is None:
        tag = default_tag
    try:
        vpc = ec2.Vpc(vpcid)
        sgs = vpc.security_groups.all()
        sg = None
        for i in sgs:
            log.debug("sg name {}".format(i.group_name))
            if "default" in i.group_name:
                sg = i
                break
        if sg == None:
            log.info("No default named security group")
            return None
    except Exception as error:
        log.info("default sg get error: {}".format(str(error)))
        return None
    try:
        #sg = vpc.create_security_group(
        #    Description='virtqe s1',
        #    GroupName='default',
        #    DryRun=True
        #)
        #log.info("New security group created %s", sg.id)
        sg.create_tags(
            DryRun=False,
            Tags=[
                {
                    'Key': 'Name',
                    'Value': tag
                },
            ]
        )
        log.info("tag added")
        response = sg.authorize_ingress(
            IpPermissions=[
                {
                    "PrefixListIds": [],
                    "FromPort": 22,
                    "IpRanges": [
                        {
                            "CidrIp": "0.0.0.0/0"
                        }
                    ],
                    "ToPort": 22,
                    "IpProtocol": "tcp",
                    "UserIdGroupPairs": [],
                    "Ipv6Ranges": []
                },
                {
                    "PrefixListIds": [],
                    "FromPort": -1,
                    "IpRanges": [],
                    "ToPort": -1,
                    "IpProtocol": "icmpv6",
                    "UserIdGroupPairs": [],
                    "Ipv6Ranges": [
                        {
                            "CidrIpv6": "::/0"
                        }
                    ]
                },
                {
                    "PrefixListIds": [],
                    "FromPort": -1,
                    "IpRanges": [
                        {
                            "CidrIp": "0.0.0.0/0"
                        }
                    ],
                    "ToPort": -1,
                    "IpProtocol": "icmp",
                    "UserIdGroupPairs": [],
                    "Ipv6Ranges": []
                }
            ]
        )
        log.info("Enabled ssh port created {}".format(sg.id))
        return sg
    except Exception as err:
        log.info(str(err))
        return None

def aws_subnet_create(region, profile, vpcid, tag, log=None):
    '''
    create a new subnet
    '''
    if log is None:
        log = minilog.minilog()
    ec2, client = aws_init_key(region,profile=profile, log=log)
    default_tag = 'virtqe'
    if tag is None:
        tag = default_tag
    try:
        vpc = ec2.Vpc(vpcid)
        subnet = vpc.create_subnet(
            CidrBlock='192.111.1.0/24',
            DryRun=False
        )
        log.info("New subnet created {}".format(subnet.id))
        subnet.create_tags(
            DryRun=False,
            Tags=[
                {
                    'Key': 'Name',
                    'Value': tag
                },
            ]
        )
        log.info("tag added")
        client.modify_subnet_attribute(
            MapPublicIpOnLaunch={
                'Value': True
            },
            SubnetId=subnet.id
        )
        log.info("enabled ipv4 on launch")
        return subnet
    except Exception as err:
        log.info(str(err))
        return None

def aws_import_key(region=None, profile=None, keyname=None, pubkeyfile=None, log=None):
    '''
    if keyname does not exists, try to import it.
    '''
    if log is None:
        log = minilog.minilog()
    ec2_resource, ec2_client = aws_init_key(region, profile, log=log)
    try:
        keypair = ec2_resource.KeyPair(keyname)
        keypair.key_pair_id
        log.info("{} exists.".format(keyname))
        return True
    except Exception as exc:
        log.info("{} not found with error{}".format(exc))

    pubkeystr = None
    if not os.path.exists(pubkeyfile):
        log.info("{} not found, do not try to import local key".format(pubkeyfile))
        return False
    with open(pubkeyfile, 'r') as fh:
        pubkeystr=fh.readlines()[0]
    try:
        response = ec2_client.import_key_pair(
            DryRun=False,
            KeyName=keyname,
            PublicKeyMaterial=pubkeystr
        )
        log.info("{} added!".format(keyname))
    except Exception as error:
        if 'Duplicate' in str(error):
            log.info("{} already exists in {}".format(keyname, region))
        else:
            log.info(error)

def aws_find_region_missed(profile=None,resource_file=None, log=None):
    if log is None:
        log = minilog.minilog()
    ec2_resource, ec2_client = aws_init_key(profile=profile, log=log)
    region_list = ec2_client.describe_regions()['Regions']
    region_name_list = [region['RegionName'] for region in region_list]
    missed_list = copy.deepcopy(region_name_list)
    has_list = []
    with open(resource_file,'r') as fh:
        tmp_resource = json.load(fh)
        for r in tmp_resource:
            if r['region'] in region_name_list and r['region'] in missed_list:
                missed_list.remove(r['region'])
            if r['region'] not in has_list:
                has_list.append(r['region'])
    return missed_list, has_list

    
def aws_check_all_regions(profile=None, is_paralle=True, log=None, resource_file=None):
    '''
    this func checks all regions has keypair required.
    and search subnet, sg to allow to create instance and make ssh connection
    '''
    # Config file
    cfg_file = os.path.dirname(ami_val.__file__) + "/cfg/ami-val.yaml"
    # Result dir
    with open(cfg_file,'r') as fh:
       keys_data = load(fh, Loader=Loader)
    if log is None:
        log = minilog.minilog()
    keyname, pubkeyfile = keys_data['pair_name'], keys_data['ssh_pubfile']
    ec2_resource, ec2_client = aws_init_key(profile=profile, log=log)
    region_list = ec2_client.describe_regions()['Regions']
    log.info("Checking keypair exists in all regions......")
    if is_paralle:
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            all_jobs = {executor.submit(aws_import_key, region['RegionName'], profile, keyname, pubkeyfile): region for region in region_list}
            for r in concurrent.futures.as_completed(all_jobs):
                x = all_jobs[r]
                try:
                    data = r.result()
                except Exception as exc:
                    print("{} generated an exception: {}".format(r,exc))
                else:
                    pass
    else:
        for region in region_list:
            aws_import_key(region=region['RegionName'], profile=profile, keyname=keyname, pubkeyfile=pubkeyfile)
    log.info("Searching proper subnet and security group in all regions......")
    if is_paralle:
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            all_jobs = {executor.submit(aws_subnet_find, region['RegionName'], profile=profile, create_new=False): region for region in region_list}
            for r in concurrent.futures.as_completed(all_jobs):
                x = all_jobs[r]
                try:
                    data = r.result()
                    log.info(r.result())
                    utils_lib.save_resource(resource_file, region=data[2], subnet=data[0], sg=data[1])
                except Exception as exc:
                    print("{} generated an exception: {}".format(r,exc))
                else:
                    pass
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            all_jobs = {executor.submit(aws_instance_type_find, region['RegionName'], profile=profile, arch='arm64'): region for region in region_list}
            for r in concurrent.futures.as_completed(all_jobs):
                x = all_jobs[r]
                try:
                    data = r.result()
                    log.info(r.result())
                    utils_lib.save_resource(resource_file, region=data[0], arch='arm64', instance_types_list=data[1])
                except Exception as exc:
                    print("{} generated an exception: {}".format(r,exc))
                else:
                    pass
        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            all_jobs = {executor.submit(aws_instance_type_find, region['RegionName'], profile=profile, arch='x86_64'): region for region in region_list}
            for r in concurrent.futures.as_completed(all_jobs):
                x = all_jobs[r]
                try:
                    data = r.result()
                    log.info(r.result())
                    utils_lib.save_resource(resource_file, region=data[0], arch='x86_64', instance_types_list=data[1])
                except Exception as exc:
                    print("{} generated an exception: {}".format(r,exc))
                else:
                    pass
    
    else:
        for region in region_list:
            subnet_id, sg_id, region_name = aws_subnet_find(region=region['RegionName'], profile=profile, create_new=False)
            #log.info('{} {} {}'.format(subnet_id, sg_id, region_name))
            utils_lib.save_resource(resource_file, region=region_name, subnet=subnet_id, sg=sg_id)
            _, instance_type_arm = aws_instance_type_find(region=region['RegionName'], profile=profile, arch='arm64')
            _, instance_type_x86 = aws_instance_type_find(region=region['RegionName'], profile=profile, arch='x86_64')
            utils_lib.save_resource(resource_file, region=region['RegionName'], arch='arm64', instance_types_list=instance_type_arm)
            utils_lib.save_resource(resource_file, region=region['RegionName'], arch='x86_64', instance_types_list=instance_type_x86)


def aws_check_region(region=None, profile=None, log=None, resource_file=None):
    '''
    this func checks all regions has keypair required.
    and search subnet, sg to allow to create instance and make ssh connection
    '''
    # Config file
    cfg_file = os.path.dirname(ami_val.__file__) + "/cfg/ami-val.yaml"
    # Result dir
    with open(cfg_file,'r') as fh:
       keys_data = load(fh, Loader=Loader)
    if log is None:
        log = minilog.minilog()
    keyname, pubkeyfile = keys_data['pair_name'], keys_data['ssh_pubfile']
    ec2_resource, ec2_client = aws_init_key(profile=profile, log=log)
    log.info("Checking keypair exists in {}......".format(region))
    aws_import_key(region=region, profile=profile, keyname=keyname, pubkeyfile=pubkeyfile, log=log)
    log.info("Searching proper subnet and security group in {}......".format(region))
    subnet_id, sg_id, region_name = aws_subnet_find(region=region, profile=profile, create_new=False, log=log)
    #log.info('{} {} {}'.format(subnet_id, sg_id, region_name))
    utils_lib.save_resource(resource_file, region=region, subnet=subnet_id, sg=sg_id, log=log)
    _, instance_type_arm = aws_instance_type_find(region=region, profile=profile, arch='arm64', log=log)
    _, instance_type_x86 = aws_instance_type_find(region=region, profile=profile, arch='x86_64', log=log)
    utils_lib.save_resource(resource_file, region=region, arch='arm64', instance_types_list=instance_type_arm, log=log)
    utils_lib.save_resource(resource_file, region=region, arch='x86_64', instance_types_list=instance_type_x86, log=log)

def aws_instance_type_find(region=None, profile=None, arch=None, skip_metal=True, max_mem=32, log=None):
    '''
    arch: arm64 | i386 | x86_64
    '''
    if log is None:
        log=minilog.minilog()
    _, ec2_client = aws_init_key(region=region, profile=profile, log=log)
    instance_types_list = []
    filters = []
    if arch is not None:
        filters.append({
                'Name': 'processor-info.supported-architecture',
                'Values': [
                    arch,
                ]
            })
    if skip_metal:
        filters.append({
                'Name': 'bare-metal',
                'Values': [
                    'false',
                ]
            })
    filters.append({
                'Name': 'memory-info.size-in-mib',
                'Values': [
                    str(max_mem*1024),
                ]
            })
    
    tmp_dict_all = ec2_client.describe_instance_types(Filters=filters)
    i = 0
    while True:
        #log.info('Get loop {}'.format(i))
        i = i + 1
        tmp_dict = tmp_dict_all['InstanceTypes']
        instance_types_list.extend(tmp_dict)
        try:
            nexttoken = tmp_dict_all['NextToken']
        except KeyError as err:
            log.info("Get instance types done, length {}".format(len(instance_types_list)))
            break
        if nexttoken == None:
            log.info("Get instance types done, length {}".format(len(instance_types_list)))
            break
        tmp_dict_all = ec2_client.describe_instance_types(NextToken=nexttoken, Filters=filters)
    instance_list = [ x['InstanceType'] for x in instance_types_list ]
    return region, instance_list
    



import os
import sys
import time
import json
from ami_val.libs.utils_lib import run_cmd

def test_stage1_check_bash_history(test_instance):
    for user in ['ec2-user', 'root']:
        cmd = 'sudo cat ~{}/.bash_history'.format(user)
        run_cmd(test_instance, cmd, expect_not_ret='0', msg='check bash history does not exist in fresh AMI')

def test_stage1_check_cmdline_console(test_instance):
    '''
    rhbz: 1679763 [RFE] AWS AMI - Add Amazon Timesync Service
    '''
    run_cmd(test_instance, "sudo cat /etc/chrony.conf", expect_ret=0, expect_kw='server 169.254.169.123', msg='check chrony points to Amazon Time Sync service')

def test_stage1_check_cmdline_console(test_instance):
    '''
    console output shoud be redirected to serial for for hvm instances
    '''
    run_cmd(test_instance, "sudo cat /proc/cmdline", expect_ret=0, expect_kw='console=ttyS0', msg='check serial console is redirected to ttyS0')

def test_stage1_check_cmdline_nvme_io_timeout(test_instance):
    '''
    rhbz: 1732506
    change default /sys/module/nvme_core/parameters/io_timeout from 30 to 4294967295
    '''
    run_cmd(test_instance, "sudo cat /proc/cmdline", expect_ret=0, expect_kw='nvme_core.io_timeout=4294967295', msg='checking cmdline')
    out = run_cmd(test_instance, 'sudo lsblk')
    if 'nvme' in out:
        run_cmd(test_instance, "sudo cat /sys/module/nvme_core/parameters/io_timeout", expect_ret=0, expect_kw='4294967295', msg='check actual value')

def test_stage1_check_inittab(test_instance):
    '''
    check default runlevel or systemd target
    '''
    is_systemd = run_cmd(test_instance, 'rpm -q systemd > /dev/null && echo True || echo False')
    test_instance.log.info("Is systemd system:{}".format(is_systemd))
    if 'True' in is_systemd:
        run_cmd(test_instance, "sudo readlink -f /etc/systemd/system/default.target", expect_ret=0, expect_kw='/lib/systemd/system/multi-user.target', 
            msg='check default runlevel or systemd target')
    else:
        run_cmd(test_instance, "sudo grep '^id:' /etc/inittab", expect_ret=0, expect_kw='id:3:initdefault', 
            msg='check default runlevel or systemd target')
        out = run_cmd(test_instance, 'uname -r')
        if 'el5' in out:
            run_cmd(test_instance, "grep '^si:' /etc/inittab", expect_kw="si::sysinit:/etc/rc.d/rc.sysinit")

def test_stage1_check_nameserver(test_instance):
    '''
    check if DNS resolving works
    '''
    run_cmd(test_instance, "ping -c 5 google-public-dns-a.google.com", expect_ret=0, msg='check if DNS resolving works')

def test_stage1_check_pkg_signed(test_instance):
    '''
    check no pkg signature is none,
    and check that specified gpg keys are installed
    '''
    cmd = "sudo rpm -q gpg-pubkey|wc -l"
    run_cmd(test_instance, cmd, expect_ret=0, expect_kw='2', msg='check gpg-pubkey installed')
    cmd = "sudo rpm -qa --qf '%{NAME}-%{VERSION}-%{RELEASE} %{SIGPGP:pgpsig}\n'|grep -v gpg-pubkey"
    run_cmd(test_instance, cmd, expect_ret=0, expect_not_kw='none', msg='check no pkg signature is none')
    cmd = "sudo rpm -qa --qf '%{NAME}-%{VERSION}-%{RELEASE} %{SIGPGP:pgpsig}\n'|grep -v gpg-pubkey|awk -F' ' '{print $NF}'|sort|uniq|wc -l"
    run_cmd(test_instance, cmd, expect_ret=0, expect_kw='1', msg='check use only one keyid')

def test_stage1_check_rhui_pkg(test_instance):
    aminame = test_instance.info['name']
    if 'HA' in aminame:
        test_instance.log.info('HA AMI found')
        rhui_pkg = 'rh-amazon-rhui-client-ha'
    elif 'SAP' in aminame:
        test_instance.log.info('SAP AMI found')
        # rhel7.9 does not have e4s repo, but rhel7.7, 8.2 have
        # so do not check e4s as I do not know which should have either.
        rhui_pkg = 'rh-amazon-rhui-client-sap-bundle'
        #rhui_pkg = 'rh-amazon-rhui-client-sap-bundle-e4s'
    else:
        test_instance.log.info('RHEL AMI found')
        rhui_pkg = 'rh-amazon-rhui-client'
    cmd = 'sudo rpm -qa|grep rhui'
    run_cmd(test_instance, cmd, expect_ret=0, expect_kw=rhui_pkg,msg='get rhui pkg version')

def test_stage1_check_root_is_locked(test_instance):
    """
    Root account should be locked
    """
    aminame = test_instance.info['name']
    if 'ATOMIC' in aminame.upper():
        test_instance.log.info('Atomic AMI found')
        cmd = 'sudo passwd -S root | grep -q Alternate'
    else:
        test_instance.log.info('RHEL AMI found')
        cmd = 'sudo passwd -S root | grep -q LK'
    run_cmd(test_instance, cmd, expect_ret=0, msg='check root is locked')

def test_stage1_check_shells(test_instance):
    """
    Check for bash/nologin shells in /etc/shells
    """
    run_cmd(test_instance, 'sudo cat /etc/shells', expect_kw='/bin/bash', msg='check /bin/bash in /etc/shells')

def test_stage1_check_timezone(test_instance):
    '''
    rhbz: 1187669
    check that the default timezone is set to UTC
    '''
    run_cmd(test_instance, 'date', expect_kw='UTC', msg='check timezone is set to UTC')

def test_stage1_check_username(test_instance):
    for user in ['fedora', 'cloud-user']:
        cmd = 'sudo cat /etc/passwd|grep {}'.format(user)
        run_cmd(test_instance, cmd, expect_not_ret='0', msg='check no {} user in fresh AMI'.format(user))
    run_cmd(test_instance, 'whoami', expect_kw=test_instance.vm.ssh_username, msg='check default user')
    run_cmd(test_instance, 'uname -a', msg='get kernel version')


import setuptools
import os

setuptools.setup(
    name="ami-val",
    version="0.0.1",
    author="Xiao Liang",
    author_email="xiliang@redhat.com",
    description="AMI validation tool",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/liangxiao1/ami-val",
    #packages=setuptools.find_packages(),
    packages=[ 'ami_val', 'ami_val.tests', 'ami_val.libs'],
    package_data={
        'ami_val': [
            'data/*',
            'cfg/*',
            'scripts/*'
        ]
    },
    include_package_data=True,
    #data_files=[('/'+os.path.expanduser("~"), ['cfg/os-tests.yaml']),],
    install_requires=['PyYAML', 'argparse', 'tipset', 'filelock', 'awscli', 'boto3'],
    license="GPLv3+",
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Programming Language :: Python :: 3",
        'Operating System :: POSIX',

    ],
    python_requires='>=3.6',
    entry_points = {
             'console_scripts': [
                 'ami-val = ami_val.ami_val:main',
             ],
         },
)



import unittest
import argparse
import copy
import os
import shutil
import sys
from ami_val import tests
from ami_val.libs import utils_lib, resource_class, aws_lib
from ami_val.libs.resource_class import FailException, SkipException
import ami_val
import json
import concurrent.futures
import signal
import traceback

def sig_handler(signum, frame):
    print('Got signal {}, call cleanup and exit!'.format(signum))
    #cleanup_paralle()
    sys.exit(0)

def cleanup_paralle(ts_list):
    with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
        all_jobs = {executor.submit(cleanup_single, ts): ts for ts in ts_list}
        for r in concurrent.futures.as_completed(all_jobs):
            x = all_jobs[r]
            try:
                data = r.result()
            except Exception as exc:
                print("{} generated an exception: {}".format(r,exc))
            else:
                pass

def cleanup_single(ts):
    try:
        if ts.vm is not None:
            ts.vm.delete(wait=False)
    except Exception as error:
        pass

def init_ts_vm(ts, instance_id=None):
    vm = resource_class.EC2VM(ts)
    if instance_id is None and vm.create():
        ts.vm = vm
        ts.ssh_client = vm.new_ssh_client()
    else:
        if vm.reuse_init(instance_id):
            ts.vm = vm
        #ts.ssh_client = vm.new_ssh_client()
    return ts

def main():
    parser = argparse.ArgumentParser(
    description="ami-val is a lightweight, fast check and tests collection for AMIs.")
    parser.add_argument('-l', dest='is_listcase', action='store_true',
                    help='list supported cases without run', required=False)
    parser.add_argument('-p', dest='pattern', default=None, action='store',
                    help='filter case by name', required=False)
    parser.add_argument('-s', dest='skip_pattern', default=None, action='store',
                    help='skip cases', required=False)
    parser.add_argument('-f', dest='amis_file', default=None, action='store',
                    help='Amis json file or the task url', required=False)
    parser.add_argument('--logdir', dest='logdir', default='/tmp/ami_val', action='store',
                    help='specify logdir, default is /tmp/ami_val if only amis json file specified', required=False)
    parser.add_argument('--clean', dest='is_clean', action='store_true', default=False,
                    help='release resource created in logdir specified', required=False)
    parser.add_argument('--profile', dest='profile', default='default', action='store',
                    help='aws credential profile name, default is default', required=False)
    parser.add_argument('--paralle', dest='is_paralle', action='store_true', default=False,
                    help='Run tests in all regions in paralle', required=False)
    args = parser.parse_args()

    signal.signal(signal.SIGHUP, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGQUIT, sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    print("Run in mode: is_listcase:{} pattern:{} is_paralle:{}".format(args.is_listcase, args.pattern, args.is_paralle))
    found_cases = utils_lib.tests_discover(args)
    if args.is_listcase:
        i = 1
        for case in found_cases:
            print("Case: {} ({}/{})".format(case, i, len(found_cases)))
            i+=1
        print("Total case num: %s"%len(found_cases))
        sys.exit(0)
    if args.is_clean:
        logdir = args.logdir
        if logdir is None:
            print("Please specify --logdir if you want to do cleanup")
            sys.exit(0)
        amis_file = logdir + '/images.json'
    else:
        logdir, amis_file = utils_lib.prepare_dir(args)
    sum_log = "{}/sum.json".format(logdir)
    resource_file = "{}/resource.json".format(logdir)
    if not args.is_clean:
        shutil.copy(amis_file, resource_file )
    ec2_profile = 'default'
    all_ts = []
    with open(amis_file) as fh:
        amis_dict = json.load(fh)
        for ami in amis_dict:
            ts = resource_class.BaseTest()
            ts.info = ami
            ts.logdir = logdir
            ts.sumlog = sum_log
            if ts.info['region'].startswith('us-gov'):
                ec2_profile = 'aws-us-gov'
            elif ts.info['region'].startswith('cn-'):
                ec2_profile = 'aws-china'
            else:
                ec2_profile = 'aws'
            ts.profile_name = ec2_profile
            ts.resource_file = resource_file
            all_ts.append(ts)

    print("Use profile:{}".format(ec2_profile))
    print("resource {}".format(resource_file))
    if args.is_clean:
        with open(resource_file,'r') as fh:
            resource = json.load(fh)
            for r in resource:
                if 'instance_id' in r.keys():
                    instance_id = r['instance_id']
                    region = r['region']
                    ami_id = r['ami']
                    for ts in all_ts:
                        if ts.info['ami'] == ami_id:
                            init_ts_vm(ts, instance_id=instance_id)
                            if ts.vm is not None:
                                ts.vm.delete(wait=False)
        sys.exit(0)
    #aws_lib.aws_check_all_regions(profile=ec2_profile, is_paralle=args.is_paralle, resource_file=resource_file)
    region_missed, region_uploaded = aws_lib.aws_find_region_missed(profile=ec2_profile, resource_file=resource_file)
    if not os.path.exists(sum_log):
        with open(sum_log, 'w+') as fh:
            json.dump([], fh, indent=4)
    i = 1

    for case in found_cases:
        print("Running {} ({}/{})".format(case, i, len(found_cases)))
        case_run = eval(case)
        i+=1
        for ts in all_ts:
            ts.casename = case
            ts.debuglog = "{}/debug/{}_{}.log".format(logdir, ts.info["ami"], case)
            ts.log.logfile = ts.debuglog

        # run test in paralle or sequence
        if args.is_paralle:
            with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
                all_jobs = {executor.submit(case_run, ts): ts for ts in all_ts if 'stage0' in ts.casename or ts.vm is not None}
                for r in concurrent.futures.as_completed(all_jobs):
                    ts = all_jobs[r]
                    try:
                        data = r.result()
                    except FailException as exc:
                        pass
                    except SkipException as exc:
                        pass
                    except Exception as exc:
                        #traceback.print_exc()
                        #print("{} generated an exception: {}".format(r,exc))
                        ts.fail(traceback.format_exc(), is_raise=False)
                    else:
                        pass
        else:
            for ts in all_ts:
                if ts.vm is None and 'stage0' not in ts.casename:
                    ts.skipTest("Cannot create instance in {}".format(ts.info['region'], is_raise=False ))
                    continue
                if ts.ssh_client is None and 'stage0' not in ts.casename:
                    ts.skipTest("Cannot make ssh connection in {}".format(ts.info['region'], is_raise=False))
                    continue
                try:
                    case_run(ts)
                except FailException as exc:
                    pass
                except SkipException as exc:
                    pass
                except Exception as exc:
                    ts.fail(traceback.format_exc(), is_raise=False)
        for ts in all_ts:
            ts.load_resource()
            if 'instance_types_list' in ts.resource_info.keys() and len(ts.resource_info['instance_types_list']) == 0:
                ts.skipTest("No supported instance in {}".format(ts.info['region']), is_raise=False)
                continue
            if ts.vm is None and 'instance_types_list' in ts.resource_info.keys():
                ts.skipTest("Cannot create instance in {}".format(ts.info['region']), is_raise=False)
                continue
            if ts.vm is None and 'instance_types_list' in ts.resource_info.keys():
                ts.skipTest("Cannot create instance in {}".format(ts.info['region']), is_raise=False)
                continue
            if ts.ssh_client is None and 'instance_types_list' in ts.resource_info.keys():
                ts.skipTest("Cannot make ssh connection in {}".format(ts.info['region']), is_raise=False)
                continue
        # save tests result
        for ts in all_ts:
            ts.log.info('Finish {} {} {} result:{}'.format(ts.info["region"], ts.info["ami"], ts.casename, ts.result))
            ts.sum_log(result=ts.result, error=ts.error)
            #print("{}:{}:{}".format(ts.info['ami'],ts.casename,ts.result))
            ts.result = 'PASS'
            ts.error = ''
    #cleanup(all_ts)
    print("Total case num:{} Tested AMIs:{}".format(len(found_cases), len(all_ts)))
    print("Log dir:{}".format(logdir))
    utils_lib.write_html(sum_log, region_missed=region_missed, region_uploaded=region_uploaded)
    print("HTML summary:{}/sum.html".format(logdir))
    print("Please wait resource cleanup done......")
    cleanup_paralle(all_ts)

if __name__ == "__main__":
    main()
